#[macro_use]
extern crate failure;
extern crate ctap_hmac as ctap;
#[macro_use]
extern crate serde_derive;
#[macro_use]
extern crate pamsm;
use crate::cli_args::{CommaSeparated, HexEncoded};
use crate::device::*;
use crate::error::*;
use crate::luks::*;
use ctap::FidoCredential;
use failure::_core::time::Duration;
use pamsm::PamLibExt;
use pamsm::*;
use std::collections::{HashMap, HashSet};
use std::ffi::CStr;
use std::path::Path;
use std::str::FromStr;
use sudo::{self, RunningAs};

pub mod cli_args;
pub mod device;
pub mod error;
pub mod luks;
pub mod util;

struct PamFido2Luks;

impl PamFido2Luks {
    fn open(
        &self,
        user: String,
        mut password: impl FnMut() -> PamResult<String>,
        args: Vec<String>,
    ) -> Fido2LuksResult<()> {
        let args: HashMap<String, String> = args
            .into_iter()
            .filter_map(|arg| {
                let mut parts = arg.split("=");
                parts
                    .by_ref()
                    .next()
                    .map(|key| (key.to_string(), parts.collect::<Vec<_>>().join("=")))
            })
            .collect();

        let credentials = args
            .get("credentials")
            .map(|creds| {
                <CommaSeparated<String>>::from_str(creds)
                    .expect("Invalid credentials")
                    .0 //TODO: proper error handling
            })
            .unwrap_or_default();
        let pin = args.get("pin");
        let device = args
            .get("device")
            .map(|device| device.replace("%user%", user.as_str()));
        let name = args
            .get("name")
            .map(|name| name.replace("%user%", user.as_str()));

        let attempts = args
            .get("attempts")
            .and_then(|a| a.parse::<usize>().ok())
            .unwrap_or(3);

        if let (Some(device), Some(name)) = (device, name) {
            if !Path::new(&device).exists() || Path::new(&format!("/dev/mapper/{}", name)).exists()
            {
                return Ok(());
            }
            // root required to mount luks
            match sudo::check() {
                RunningAs::User => {
                    //err
                    unimplemented!("no root")
                }
                _ => {
                    sudo::escalate_if_needed().unwrap();
                }
            }
            let mut device = LuksDevice::load(device)?;
            let mut additional_credentials: HashSet<String> = HashSet::new();
            if device.is_luks2()? {
                for token in device.tokens()? {
                    let (_, token) = token?;
                    additional_credentials.extend(token.credential.into_iter());
                }
            }
            let credentials: Vec<FidoCredential> = credentials
                .into_iter()
                .chain(additional_credentials.into_iter())
                .map(|cred| HexEncoded::from_str(cred.as_str()))
                .map(|cred| FidoCredential {
                    id: cred.unwrap().0,
                    public_key: None,
                })
                .collect();
            let credentials: Vec<&FidoCredential> = credentials.iter().collect();
            if !credentials.is_empty() {
                for _ in 0..attempts {
                    let salt = util::sha256(&[password().expect("Password").as_bytes()]);
                    let secret = util::sha256(&[
                        &salt,
                        &perform_challenge(
                            &credentials[..],
                            &salt,
                            Duration::from_secs(15),
                            pin.map(AsRef::as_ref),
                        )?
                        .0[..],
                    ]);
                    device.activate(name.as_str(), &secret[..], None)?;
                }
                Ok(())
            } else {
                unimplemented!("custom error")
            }
        } else {
            unimplemented!("custom error")
        }
    }
}

impl PamServiceModule for PamFido2Luks {
    fn authenticate(pamh: Pam, flag: PamFlag, args: Vec<String>) -> PamError {
        let user = match pamh.get_cached_user() {
            Err(_) => return dbg!(PamError::AUTH_ERR),
            Ok(p) => p.map(|s| s.to_str().map(str::to_string).unwrap()),
        };
        let mut password = match pamh.get_authtok(None) {
            Err(_) => return dbg!(PamError::AUTH_ERR),
            Ok(p) => p.map(|s| s.to_str().map(str::to_string).unwrap()),
        };
        if let Some(user) = user {
            match PamFido2Luks.open(
                user,
                move || match password.take() {
                    Some(pass) => Ok(pass),
                    None => pamh
                        .conv(Some("Fido2 salt: "), PamMsgStyle::PROMPT_ECHO_OFF)
                        .map(|s| s.map(|s| s.to_str().unwrap()).unwrap_or("").to_string()),
                },
                args,
            ) {
                Ok(_) => PamError::SUCCESS,
                Err(e) => match e {
                    //TODO: output more detailed error
                    _ => PamError::AUTH_ERR,
                },
            }
        } else {
            PamError::AUTH_ERR
        }
    }
}

pam_module!(PamFido2Luks);
